# mpf-monitor
__version__ = '0.57.2.dev2'
__short_version__ = '0.57'
__bcp_version__ = '1.1'
__config_version__ = '6'

version = f"MPF Monitor v{__version__} (config_version={__config_version__}, BCP v{__bcp_version__})"
